@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
@@ -6,9 +6,10 @@ import os
 import pwd
 import sys
 from collections import namedtuple
+from contextlib import suppress
 
 appname = 'kitty'
-version = (0, 13, 1)
+version = (0, 15, 0)
 str_version = '.'.join(map(str, version))
 _plat = sys.platform.lower()
 is_macos = 'darwin' in _plat
@@ -22,13 +23,16 @@ WindowGeometry = namedtuple('WindowGeometry', 'left top right bottom xnum ynum')
 def kitty_exe():
     ans = getattr(kitty_exe, 'ans', None)
     if ans is None:
-        rpath = getattr(sys, 'bundle_exe_dir', None)
+        rpath = sys._xoptions.get('bundle_exe_dir')
         if not rpath:
-            items = frozenset(os.environ['PATH'].split(os.pathsep))
+            items = filter(None, os.environ.get('PATH', '').split(os.pathsep))
+            seen = set()
             for candidate in items:
-                if os.access(os.path.join(candidate, 'kitty'), os.X_OK):
-                    rpath = candidate
-                    break
+                if candidate not in seen:
+                    seen.add(candidate)
+                    if os.access(os.path.join(candidate, 'kitty'), os.X_OK):
+                        rpath = candidate
+                        break
             else:
                 rpath = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'launcher')
         ans = kitty_exe.ans = os.path.join(rpath, 'kitty')
@@ -45,9 +49,8 @@ def _get_config_dir():
     locations.append(os.path.expanduser('~/.config'))
     if is_macos:
         locations.append(os.path.expanduser('~/Library/Preferences'))
-    if 'XDG_CONFIG_DIRS' in os.environ:
-        for loc in os.environ['XDG_CONFIG_DIRS'].split(os.pathsep):
-            locations.append(os.path.abspath(os.path.expanduser(loc)))
+    for loc in filter(None, os.environ.get('XDG_CONFIG_DIRS', '').split(os.pathsep)):
+        locations.append(os.path.abspath(os.path.expanduser(loc)))
     for loc in locations:
         if loc:
             q = os.path.join(loc, appname)
@@ -67,10 +70,8 @@ def _get_config_dir():
 
         def cleanup():
             import shutil
-            try:
+            with suppress(Exception):
                 shutil.rmtree(ans)
-            except Exception:
-                pass
         atexit.register(cleanup)
     return ans
 
@@ -116,14 +117,13 @@ def wakeup():
 base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
 terminfo_dir = os.path.join(base_dir, 'terminfo')
 logo_data_file = os.path.join(base_dir, 'logo', 'kitty.rgba')
+logo_png_file = os.path.join(base_dir, 'logo', 'kitty.png')
 beam_cursor_data_file = os.path.join(base_dir, 'logo', 'beam-cursor.png')
 try:
     shell_path = pwd.getpwuid(os.geteuid()).pw_shell or '/bin/sh'
 except KeyError:
-    try:
+    with suppress(Exception):
         print('Failed to read login shell via getpwuid() for current user, falling back to /bin/sh', file=sys.stderr)
-    except Exception:
-        pass
     shell_path = '/bin/sh'
 
 
@@ -131,9 +131,47 @@ def glfw_path(module):
     return os.path.join(base, 'glfw-{}.so'.format(module))
 
 
-is_wayland = False
-if os.environ.get('WAYLAND_DISPLAY') and 'KITTY_ENABLE_WAYLAND' in os.environ and os.path.exists(glfw_path('wayland')):
-    is_wayland = True
+def detect_if_wayland_ok():
+    if 'WAYLAND_DISPLAY' not in os.environ:
+        return False
+    if 'KITTY_DISABLE_WAYLAND' in os.environ:
+        return False
+    wayland = glfw_path('wayland')
+    if not os.path.exists(wayland):
+        return False
+    # GNOME does not support xdg-decorations
+    # https://gitlab.gnome.org/GNOME/mutter/issues/217
+    import ctypes
+    lib = ctypes.CDLL(wayland)
+    check = lib.glfwWaylandCheckForServerSideDecorations
+    check.restype = ctypes.c_char_p
+    check.argtypes = ()
+    try:
+        ans = bytes(check())
+    except Exception:
+        return False
+    if ans == b'NO':
+        print(
+                'Your Wayland compositor does not support server side window decorations,'
+                ' disabling Wayland. You can force Wayland support using the'
+                ' linux_display_server option in kitty.conf'
+                ' See https://drewdevault.com/2018/01/27/Sway-and-client-side-decorations.html'
+                ' for more information.',
+                file=sys.stderr)
+    return ans == b'YES'
+
+
+def is_wayland(opts=None):
+    if is_macos:
+        return False
+    if opts is None:
+        return is_wayland.ans
+    if opts.linux_display_server == 'auto':
+        ans = detect_if_wayland_ok()
+    else:
+        ans = opts.linux_display_server == 'wayland'
+    setattr(is_wayland, 'ans', ans)
+    return ans
 
 
 supports_primary_selection = not is_macos
